/* Copyright (c) 2006 Rice University */
/* All Rights Reserved */
/* This code is covered by the Rice-WARP license */
/* See http://warp.rice.edu/license/ for details */


#include "xmk.h"
#include <string.h>
#include <stdlib.h>
#include <os_config.h>
#include <sys/process.h>
#include <pthread.h>
#include <semaphore.h>
#include "xparameters.h"
#include "stdio.h"
#include "xutil.h"
#include "warplib.h"
#include "xgpio_l.h"
#include "xuartlite_l.h"


//========================================================================
// This system has been built to demonstrate the use of semaphores in
// xilkernel. The semaphore functions are used for inter thread/process
// management and control. A small example is that semaphores can be used
// to pause a process of any priority and jump to another thread. The
// concept relies on the use of a locking mechanism and the sem_wait()
// function -- this function will cause a function to wait until
// a semaphore has been released by a different function.
//
// Oh yeah, you could tie processes to specific peripherals and use
// the semaphores to effectively control peripheral access.
//
// This demo will create a master thread to control the others via semaphores.
// The other threads will perform some computations on a global variable,
// demonstrating a basic example of ORDER OF OPERATIONS. The global variable
// is a light replacement for a shared memory buffer -- it's probably
// not good programming practice to do this (and requires a reinitialization
// even after system restart!). Shared memory is for another time though...
//
//
//
// Written by:	Scott Novich (snovich@rice.edu)
// Portions of this code refer to Xilinx's primary xilkernel XPS demo
// for Nokia and Rice
//
//========================================================================

// the following declarations are neccessary for thread creation
static pthread_t athread, mthread; // identifiers for creating new pthreads
static pthread_attr_t threadattr; // an attribute identifer used for creating the new pthread
static struct sched_param newthreadsched; // used for giving our new thread a priority level
static sem_t MasterSemaphore, SubSem1, SubSem2; //  semaphore name instantiations
int GLOBAL_ANSWER=1; // this is a weak-kind of way of instantiating "shared memory." OK for this demo.
int genericvalue;
static void *retptr; //used for join function
static int ret; // some return holder

void* adderthread(void *arg1)
{
	// at program init, SubSem1 was locked
	// so sem_wait will cause this thread to freeze
	// until INITCTRL releases it's lock using sem_post
	sem_wait(&SubSem1);

	// this semaphore was UNLOCKED
	// at INITCTRL start. So whatever thread
	// gets to this first, has the lock.
	// The effect of this is that whichever
	// sem_post got called first in INITCTRL,
	// will cause the other thread to lock up
	// so both adderthread and multthread
	// cannot run simultaneously, despite
	// having the same priority level
	// which would typically cause the PPC
	// to process both simultaneously in RR style.
	sem_wait(&MasterSemaphore);

	// NOW onto the real algorithm

	GLOBAL_ANSWER=GLOBAL_ANSWER+10; // INTENSE!
	sleep(100); // add a small delay for no reason

	// release the master-lock!
	sem_post(&MasterSemaphore);
	sem_post(&SubSem1);
	pthread_exit(&ret);
}
void* multthread(void *arg2)
{
	// at program init, SubSem1 was locked
	// so sem_wait will cause this thread to freeze
	// until INITCTRL releases it's lock using sem_post
	sem_wait(&SubSem2);

	// this semaphore was UNLOCKED
	// at INITCTRL start. So whatever thread
	// gets to this first, has the lock.
	// The effect of this is that whichever
	// sem_post got called first in INITCTRL,
	// will cause the other thread to lock up
	// so both adderthread and multthread
	// cannot run simultaneously, despite
	// having the same priority level
	// which would typically cause the PPC
	// to process both simultaneously in RR style.
	sem_wait(&MasterSemaphore);

	// NOW onto the real algorithm

	GLOBAL_ANSWER=GLOBAL_ANSWER*5; // INTENSE!
	sleep(100); // add a small delay for no reason

	// release the master-lock!
	sem_post(&MasterSemaphore);
	sem_post(&SubSem2);
	pthread_exit(&ret);
}

// we need to utilize an initial thread from xilkernel launch
// this is defined in the static pthread table under software platform settings.
// This thread will launch the other threads and instantiate the semaphores
void* INITCTRL(void *notused)
{
	print("\r\n-- Entering Initial Control Thread --\r\n");
	print("\r\n-- Generating Semaphores --");
	char method;
	int check=0;

	// This semaphore is initialized to an unlocked state
	if(sem_init(&MasterSemaphore, 1, 1) < 0)
	{
		print("\r\n MasterSemaphore init error...\r\n");
		goto terminate;
	}

	// We can also initialize semaphores to LOCKED states.
	// This thread (INITCTRL) will hold the lock on these.
	// It's useful for freezing the processing of other threads.
	if(sem_init(&SubSem1, 1, 0) < 0)
	{
		print("\r\n SubSem1 init error...\r\n");
		goto terminate;
	}
	if(sem_init(&SubSem2, 1, 0) < 0)
	{
		print("\r\n SubSem2 init error...\r\n");
		goto terminate;
	}

	// now it's time to launch the threads...see the pthread demos on how to do this.
	pthread_attr_init(&threadattr);
	newthreadsched.sched_priority = 1; // must be same priority as calling thread for this demo to work
	pthread_attr_setdetachstate(&threadattr, PTHREAD_CREATE_JOINABLE);// will be used to release resources later
	pthread_attr_setschedparam(&threadattr,&newthreadsched);

	print("\r\n-- Time to launch some threads! --");
	print("\r\n-- Starting Thread 1 --");

	genericvalue = pthread_create(&athread, &threadattr, (void*)adderthread, NULL); // NULL is a placeholder. It's used if the thread function must take in an argument
	if (genericvalue != 0)
	{
		print("\r\n Thread1 init error...\r\n");
		goto terminate;
	}
	print("\r\n-- Starting Thread 2 --");
	genericvalue = pthread_create(&mthread, &threadattr, (void*)multthread, NULL); // NULL is a placeholder. It's used if the thread function must take in an argument
	if (genericvalue != 0)
	{
		print("\r\n multthread init error...\r\n");
		goto terminate;
	}

	print("\r\n-- SYSTEM ACTIVE --");
	while(check==0)
	{
		print("\r\n Select whether you wish to add or multiply first: \r\n");
		print("(1): Add 10 first.\r\n");
		print("(2): Multiply by 5 first.\r\n");
		print("(3): Terminate Application.\r\n");
		xil_printf("\r\n Global Answer is currently: %d\r\n",GLOBAL_ANSWER);

		method = XUartLite_RecvByte(STDIN_BASEADDRESS);

		switch(method)
		{
			case('1'):
			{
				sem_post(&SubSem1); // start the add first
				sem_post(&SubSem2); // start the mult
				if(pthread_join(athread, &retptr) !=0) // join adder thread (halt exec. of this thread til it finishes);
				{
					print("\r\njoin error\r\n");
				}
				if(pthread_join(mthread, &retptr) !=0) // join mult thread (halt exec. of this thread til it finishes);
				{
					print("\r\njoin error\r\n");
				}
				check = 1;
			} break;
			case('2'):
			{
				sem_post(&SubSem2); // start the mult first
				sem_post(&SubSem1); // start the add
				if(pthread_join(mthread, &retptr) !=0)
				{
					print("\r\njoin error\r\n");
				}
				if(pthread_join(athread, &retptr) !=0)
				{
					print("\r\njoin error\r\n");
				}
				check = 1;
			}break;
			case('3'):
			{
				check = 1;
			}
			default : print("\r\n ERROR in choice. Try again.\r\n");
		}
	}
	xil_printf("\r\n Global Answer is currently: %d\r\n",GLOBAL_ANSWER);

	// from Xilinx's semaphore demo, available @ warp.rice.edu/TRAC ported for WARP v. 1.1
	if ((sem_destroy(&MasterSemaphore) != 0) || (sem_destroy(&SubSem1) !=0) || (sem_destroy(&SubSem2) !=0))
	{
		print ("SEM: Error while releasing resources.\r\n");
		goto terminate;
	}


terminate:
		 GLOBAL_ANSWER=1;
		 print("\r\n Program Terminated, going to sleep\r\n");
		 while(1)
		 {
			; // I do this so there aren't those annoying "Idle Task" RS232 prints.
		 }
}

int main () {
   print("-- Entering main() --\r\n");
	print("-- press any key to start xilkernel system--\r\n\r\n");
	//getchar();
	xilkernel_main();
	print("-- Exiting main() --\r\n");
   return 0;
}
//////////////////// END NOTES //////////////////////////////////
// the pthread_join() function must be carefully placed/
// it will only work if the target thread is currently RUNNING
// (and NOT in a semaphore wait).
/////////////////////////////////////////////////////////////////

